//License 
/***
 * Java TelnetD library (embeddable telnet daemon)
 * Copyright (c) 2000-04 Dieter Wimberger 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * Neither the name of the author nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *  
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 ***/
package net.wimpi.telnetd.net;

import java.io.IOException;
import java.net.ServerSocket;
import java.net.Socket;
import java.text.MessageFormat;
import java.util.Properties;

import net.wimpi.telnetd.BootException;
import net.wimpi.telnetd.TelnetD;

/**
 * Class that implements a PortListener.<br>
 * If listening, it accepts incoming connections, and handles them via
 * the ConnectionManager.
 *
 * @author  Dieter Wimberger
 * @version 1.0 (16/01/2004)
 *
 * @see net.wimpi.telnetd.net.ConnectionManager
 */
public class PortListener
    extends Thread {

  //Members
  private String m_Name;
  private int m_Port;							            //port number running on
  private int m_FloodProtection;					    //flooding protection
  private ServerSocket m_ServerSocket = null; //server socket
  private boolean m_Available;               		//Flag for availability


  //Associations
  private ConnectionManager m_ConnectionManager;	//connection management thread

  /**
   * Constructs a PortListener instance.<br>
   * Its private because its called by a factory method.
   *
   * @param port int that specifies the port number of the server socket.
   * @param floodprot that specifies the server socket queue size.
   */
  private PortListener(String name,int port, int floodprot) {
    m_Name = name;
    m_Available = false;
    m_Port = port;
    m_FloodProtection = floodprot;
  }//constructor


  /**
   * Overloads the parents method to listen constantly to a server
   * socket and handles incoming connections through the ConnectionManager.
   *
   * @see net.wimpi.telnetd.net.ConnectionManager
   */
  public void run() {
    try {
      /*
          A server socket is opened with a connectivity queue of a size specified
          in int floodProtection.  Concurrent login handling under normal circumstances
          should be handled properly, but denial of service attacks via massive parallel
          program logins should be prevented with this.
      */
      m_ServerSocket = new ServerSocket(m_Port, m_FloodProtection);


      //log entry
      Object[] args = {new Integer(m_Port), new Integer(m_FloodProtection)};
      TelnetD.syslog.write(MessageFormat.format(logmsg, args));

      //Infinite loop, taking care of resources in case of a lot of parallel logins
      do {
        Socket s = m_ServerSocket.accept();
        if (m_Available) {
          m_ConnectionManager.makeConnection(s);
        } else {
          //just shut down the socket
          s.close();
        }
        this.yield();
      } while (true);

    } catch (IOException e) {
      TelnetD.syslog.writeStackTrace(e);
      //FIXME: this is a major failure, we should consider
      //how to handle it (Recovery?)
      System.out.println("OHOH");
    }
    System.out.println("OHOH");

  }//run


  /**
   * Sets the state of the listening flag. If true
   * the listener will accept incoming connections,
   * if false not.
   *
   * @param b the new state of the listening flag.
   */
  public void setListening(boolean b) {
    m_Available = b;
  }//setListening

  /**
   * Returns the state of the listening flag. True
   * if the listener is accepting incoming connections,
   * false if not.
   *
   * @return the state of the listening flag.
   */
  public boolean isListening() {
    return m_Available;
  }//isListening

  /**
   * Returns reference to ConnectionManager instance associated
   * with the PortListener.
   *
   * @return the associated ConnectionManager.
   */
  public ConnectionManager getConnectionManager() {
    return m_ConnectionManager;
  }//getConnectionManager

  /**
   * Method that will shutdown the listener.
   */
  public void shutdown() {
    //take down connections
    m_ConnectionManager.shutdown();
    //stop related connectionmanager thread
    m_ConnectionManager.stop();
  }//shutdown

  /**
   * Factory method for a PortListener instance, returns
   * an instance of a PortListener with an associated ConnectionManager.
   *
   * @param settings Properties that contain all configuration information.
   */
  public static PortListener createPortListener(String name, Properties settings)
      throws BootException {

    PortListener pl = null;

    try {
      //1. read settings of the port listener itself
      int port = Integer.parseInt(settings.getProperty(name + ".port"));
      int floodprot = Integer.parseInt(settings.getProperty(name + ".floodprotection"));

      if (new Boolean(settings.getProperty(name + ".secure")).booleanValue()) {
        //do nothing for now, probably set factory in the future
      }
      pl = new PortListener(name, port, floodprot);
    } catch (Exception ex) {
      throw new BootException("Failure while creating PortListener instance:\n" +
          ex.getMessage());
    }

    //2. factorize a ConnectionManager, passing the settings, if we do not have one yet
    if (pl.m_ConnectionManager == null) {
      pl.m_ConnectionManager = ConnectionManager.createConnectionManager(name, settings);
      try {
        pl.m_ConnectionManager.start();
      } catch (Exception exc) {
        throw new BootException("Failure while starting ConnectionManager watchdog thread:\n" +
            exc.getMessage());
      }
    }

    return pl;
  }//createPortListener


//Constants
  private static final String logmsg =
      "Listening to Port {0,number,integer} with a connectivity queue size of {1,number,integer}.";

}//class PortListener
